﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using ThinkGeo.MapSuite.DesktopEdition;
using ThinkGeo.MapSuite.Core;
using System.Diagnostics;
using System.Data.SqlClient;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;

namespace DatabaseToMap
{
    public partial class Form1 : Form
    {
        private string _db = "";
        
        private string conn 
        {
            get { return @"Data Source=(local)\SQLEXPRESS_1;" + 
                "Initial Catalog=" + _db + ";" +
                "Integrated Security=True;" +
                "Connect Timeout=15;" +
                "Encrypt=False;"+
                "TrustServerCertificate=False;" + 
                "ApplicationIntent=ReadWrite;" +
                "MultiSubnetFailover=False"; }
        }

        // A type map to convert data types to db types
        Dictionary<Type, SqlDbType> typeMap = new Dictionary<Type, SqlDbType>();        

        public Form1()
        {
            InitializeComponent();

            winformsMap1.MapUnit = GeographyUnit.Meter;
            LayerOverlay lo = new LayerOverlay();
            lo.Name = "DbOverlay";
            winformsMap1.Overlays.Add("DbOverlay", lo);

            // add to the type map
            //typeMap.Add(typeof(byte), SqlDbType.Byte);
            //typeMap.Add(typeof(sbyte), SqlDbType.SByte);
            typeMap.Add(typeof(short), SqlDbType.Int);
            //typeMap.Add(typeof(ushort), SqlDbType.UInt16);
            typeMap.Add(typeof(int), SqlDbType.Int);
           // typeMap.Add(typeof(uint), SqlDbType.UInt32);
            typeMap.Add(typeof(long), SqlDbType.Int);
            //typeMap.Add(typeof(ulong), SqlDbType.UInt64);
            typeMap.Add(typeof(float), SqlDbType.Float);
            typeMap.Add(typeof(double), SqlDbType.Float);
            typeMap.Add(typeof(decimal), SqlDbType.Decimal);
            typeMap.Add(typeof(bool), SqlDbType.Bit);
            typeMap.Add(typeof(string), SqlDbType.Text);
            typeMap.Add(typeof(char), SqlDbType.NChar);
            //typeMap.Add(typeof(Guid), SqlDbType.Guid);
            typeMap.Add(typeof(DateTime), SqlDbType.DateTime);
            typeMap.Add(typeof(DateTimeOffset), SqlDbType.DateTimeOffset);
            typeMap.Add(typeof(byte[]), SqlDbType.Binary);
            //typeMap.Add(typeof(byte?), SqlDbType.Byte);
            //typeMap.Add(typeof(sbyte?), SqlDbType.SByte);
            typeMap.Add(typeof(short?), SqlDbType.Int);
            //typeMap.Add(typeof(ushort?), SqlDbType.UInt16);
            typeMap.Add(typeof(int?), SqlDbType.Int);
            //typeMap.Add(typeof(uint?), SqlDbType.UInt32);
            typeMap.Add(typeof(long?), SqlDbType.Int);
            //typeMap.Add(typeof(ulong?), SqlDbType.UInt64);
            typeMap.Add(typeof(float?), SqlDbType.Float);
            typeMap.Add(typeof(double?), SqlDbType.Float);
            typeMap.Add(typeof(decimal?), SqlDbType.Decimal);
            typeMap.Add(typeof(bool?), SqlDbType.Bit);
            typeMap.Add(typeof(char?), SqlDbType.NChar);
            //typeMap.Add(typeof(Guid?), SqlDbType.Guid);
            typeMap.Add(typeof(DateTime?), SqlDbType.DateTime);
            typeMap.Add(typeof(DateTimeOffset?), SqlDbType.DateTimeOffset);

        }

        private void btnCreate_Click(object sender, EventArgs e)
        {

            SaveFileDialog dlg = new SaveFileDialog();
            dlg.Title = "Save local db";
            dlg.Filter = "SQL Server Db (*.mdf)|*.mdf";
            dlg.InitialDirectory = @"C:\Program Files\Microsoft SQL Server\MSSQL10_50.SQLEXPRESS_1\MSSQL\DATA";

            if (dlg.ShowDialog() != DialogResult.OK) return;

            string ext = new FileInfo(dlg.FileName).Extension;
            string name = new FileInfo(dlg.FileName).Name;
            name = name.Replace(ext, "");

            SqlConnection myConn = new SqlConnection(@"Server=(local)\SQLEXPRESS_1;" +
                "Integrated Security=True;" +
                "database=master;" +
                "Connect Timeout=15;" +
                "Encrypt=False;" +
                "TrustServerCertificate=False;" +
                "ApplicationIntent=ReadWrite;" +
                "MultiSubnetFailover=False");

            string str = string.Format("CREATE DATABASE {0} ON PRIMARY " +
                "(NAME = {0}_Data, " +
                "FILENAME = '{1}', " +
                "SIZE = 2MB, MAXSIZE = 10MB, FILEGROWTH = 10%) " +
                "LOG ON (NAME = {0}_Log, " +
                "FILENAME = 'C:\\{0}Log.ldf', " +
                "SIZE = 1MB, " +
                "MAXSIZE = 5MB, " +
                "FILEGROWTH = 10%)", name, dlg.FileName);

            str = string.Format("CREATE DATABASE {0}", name);

            SqlCommand myCommand = new SqlCommand(str, myConn);
            try
            {
                myConn.Open();
                myCommand.ExecuteNonQuery();
                MessageBox.Show("DataBase is Created Successfully", "Success", 
                    MessageBoxButtons.OK, MessageBoxIcon.Information);
                _db = name;

                AddTable();
                //CreateTableFromObject("Shape", "DatabaseToMap.Shot", null);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "Epic fail", MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
            finally
            {
                if (myConn.State == ConnectionState.Open)
                {
                    myConn.Close();
                }
            }
        }

        private bool AddTable()
        {
            StringBuilder query = new StringBuilder();
            query.Append("CREATE TABLE [dbo].[Shape] (" +
                "[Id] INT identity(1,1) constraint pkid NOT NULL PRIMARY KEY," +
                "[geom] [sys].[geometry] NULL," +
                "[Name] TEXT NULL," +
                "[MyObject] NVARCHAR(MAX) NULL);");

            SqlConnection myConn = new SqlConnection(conn);

            try
            {
                myConn.Open();
                SqlCommand sqlQuery = new SqlCommand(query.ToString(), myConn);
                SqlDataReader reader = sqlQuery.ExecuteReader();

                MessageBox.Show("Table Created Successfully", "Success", 
                    MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "Epic fail", MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
            finally 
            {
                if (myConn.State == ConnectionState.Open)
                {
                    myConn.Close();
                }
            }

            return true;

        }

        private bool CreateTableFromObject(string tableName, string objectName, string foreignKey)
        {

            StringBuilder query = new StringBuilder();
            query.Append("CREATE TABLE ");
            query.Append(tableName);
            query.Append(" ( ");
            query.Append("[Id] INT IDENTITY (1, 1) NOT NULL,");


            // Get the properties to tag to each receiver
            Type myType = Type.GetType(objectName);
            foreach (var property in myType.GetProperties())
            {
                string name = property.Name;
                string typeString = "";                
                try
                {
                    SqlDbType type = typeMap[property.PropertyType];
                    typeString = type.ToString();
                }
                catch (Exception ex)
                {
                    Type baseType = property.PropertyType.BaseType;
                    while (baseType != null)
	                {
	                    baseType = baseType.BaseType;
                        if (baseType == typeof(BaseShape))
                        {
                            name = "geom";
                            typeString = "[sys].[geometry]";
                        }
	                }

                }                
                
                query.Append(name);
                query.Append(" ");
                query.Append(typeString);
                query.Append(", ");
            }

            // Append a var binary field that we can store the complete object in
            query.Append("Object");
            query.Append(" ");
            query.Append(SqlDbType.VarBinary.ToString());
            query.Append(", ");

            query.Length -= 2; //Remove trailing ", "
            
            query.Append(")");

            SqlConnection myConn = new SqlConnection(conn);

            try
            {
                myConn.Open();
                SqlCommand sqlQuery = new SqlCommand(query.ToString(), myConn);
                SqlDataReader reader = sqlQuery.ExecuteReader();

                MessageBox.Show("Table Created Successfully", "Success",
                    MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "Epic fail", MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
            finally
            {
                if (myConn.State == ConnectionState.Open)
                {
                    myConn.Close();
                }
            }

            return true;

        }

        private void ClearTable_Click(object sender, EventArgs e)
        {
            if (!CheckConn()) return;

            StringBuilder query = new StringBuilder();
            query.Append("DELETE from ");
            query.Append("Shape");
            query.Append(";");

            SqlConnection myConn = new SqlConnection(conn);

            try
            {
                myConn.Open();
                SqlCommand sqlQuery = new SqlCommand(query.ToString(), myConn);
                SqlDataReader reader = sqlQuery.ExecuteReader();

                MessageBox.Show("Rows removed Successfully", "Success",
                    MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "Epic fail", 
                    MessageBoxButtons.OK, MessageBoxIcon.Information);
            }

        }

        private void btnDelete_Click(object sender, EventArgs e)
        {
            
            if (!CheckConn()) return;

            SqlConnection myConn = new SqlConnection(@"Server=(local)\SQLEXPRESS_1;" +
                "Integrated Security=True;" +
                "database=master;" +
                "Connect Timeout=15;" +
                "Encrypt=False;" +
                "TrustServerCertificate=False;" +
                "ApplicationIntent=ReadWrite;" +
                "MultiSubnetFailover=False");

            string str = string.Format("DROP DATABASE [{0}]", _db);

            SqlCommand myCommand = new SqlCommand(str, myConn);
            try
            {
                myConn.Open();
                myCommand.ExecuteNonQuery();
                MessageBox.Show("DataBase Successfully Deleted", "Success",
                    MessageBoxButtons.OK, MessageBoxIcon.Information);

            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "Epic fail", 
                    MessageBoxButtons.OK, MessageBoxIcon.Information);
            }

        }

        private bool CheckConn()
        {
            if (_db == "")
            {
                SaveFileDialog dlg = new SaveFileDialog();
                dlg.Title = "Save local db";
                dlg.Filter = "SQL Server Db (*.mdf)|*.mdf";
                dlg.InitialDirectory = @"C:\Program Files\Microsoft SQL Server\MSSQL10_50.SQLEXPRESS_1\MSSQL\DATA";
                
                if (dlg.ShowDialog() != DialogResult.OK) return false;

                string ext = new FileInfo(dlg.FileName).Extension;
                string name = new FileInfo(dlg.FileName).Name;
                name = name.Replace(ext, "");

                _db = name;
            }

            return true;
        }

        private void btnRead_Click(object sender, EventArgs e)
        {
            Shot shot = new Shot(new PointShape(0, 0), "shot", 1, 2000);
            byte[] array = ObjectToByteArray(shot);
            string objString = System.Text.Encoding.Unicode.GetString(array);

            System.Text.UnicodeEncoding encoding = new System.Text.UnicodeEncoding();
            byte[] bytes = encoding.GetBytes(objString);
            Shot shotd = (Shot)ByteArrayToObject(bytes);

            MessageBox.Show(string.Format("Name={0}, Pressure={1}, ShotNum={2}, size={3}", shotd.Name, shotd.Pressure, shotd.ShotNum, bytes.Length));

        }

        // Convert an object to a byte array
        public static byte[] ObjectToByteArray(Object obj)
        {
            BinaryFormatter bf = new BinaryFormatter();
            using (var ms = new MemoryStream())
            {
                bf.Serialize(ms, obj);
                return ms.ToArray();
            }
        }

        // Convert a byte array to an Object
        public static Object ByteArrayToObject(byte[] arrBytes)
        {
            using (var memStream = new MemoryStream())
            {
                var binForm = new BinaryFormatter();
                memStream.Write(arrBytes, 0, arrBytes.Length);
                memStream.Seek(0, SeekOrigin.Begin);
                memStream.Position = 0;
                var obj = binForm.Deserialize(memStream);
                return obj;
            }
        }

        private void AddFeatures_Click(object sender, EventArgs e)
        {

            if (!CheckConn()) return;

            MsSql2008FeatureLayer db = new MsSql2008FeatureLayer(conn, "Shape", "Id");            
            db.Srid = 32516;
            db.CustomGeometryColumnName = "geom";   // Must set if not "geom"
            db.Open();

            var s = Stopwatch.StartNew();

            Random rnd = new Random(5);
            db.EditTools.BeginTransaction();
            int records = 1;

            try
            {
                for (int i = 0; i < records; i++)
                {

                    // Create a shot object we can serialize
                    Shot shot = new Shot(new PointShape(0, 0), "shot", i, 2000);

                    // Get byte array and encode to Unicode before saving to the db
                    byte[] array = ObjectToByteArray(shot);
                    string objString = System.Text.Encoding.Unicode.GetString(array);

                    db.EditTools.Add(new Feature(new PointShape(rnd.Next(500, 1000), rnd.Next(2000, 3000)),
                        new List<string>() { 
                        "Name:test" + i,
                        "MyObject:" + objString // Why does this not work unless it is a string???
                        }));
                    
                }              

                db.EditTools.CommitTransaction();
                MessageBox.Show("Feature count = " + db.FeatureSource.GetCount());
                db.Close();
                
            }
            catch(Exception ex)
            {
                MessageBox.Show(ex.ToString(), "Epic fail",
                    MessageBoxButtons.OK, MessageBoxIcon.Information);
            }

            MessageBox.Show(string.Format("Time to add {0} features was {1} seconds", 
                records, s.Elapsed.TotalSeconds));

        }

        private void MapFeatures_Click(object sender, EventArgs e)
        {
            if (!CheckConn()) return;

            var s = Stopwatch.StartNew();

            MsSql2008FeatureLayer db = new MsSql2008FeatureLayer(conn, "Shape", "Id");

            db.ZoomLevelSet.ZoomLevel01.DefaultPointStyle = new PointStyle(PointSymbolType.Circle, 
                new GeoSolidBrush(GeoColor.StandardColors.Red), 3);
            db.ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;
            db.Name = "test";
            if (((LayerOverlay)winformsMap1.Overlays[0]).Layers.Count > 0)
                ((LayerOverlay)winformsMap1.Overlays[0]).Layers.RemoveAt(0);
            ((LayerOverlay)winformsMap1.Overlays[0]).Layers.Add("test", db);
            
            db.Open();
            RectangleShape rect = db.GetBoundingBox();

            List<Shot> shots = new List<Shot>();
            foreach (Feature feature in db.FeatureSource.GetAllFeatures(ReturningColumnsType.AllColumns))
            {
                // Deserialize the shot and add it to the list
                string array = feature.ColumnValues["MyObject"];
                System.Text.UnicodeEncoding encoding = new System.Text.UnicodeEncoding();
                byte[] bytes = encoding.GetBytes(array);

                try
                {
                    Shot shot = (Shot)ByteArrayToObject(bytes);
                    shots.Add(shot);
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.ToString(), "Epic fail",
                        MessageBoxButtons.OK, MessageBoxIcon.Information);
                }
            }

            db.Close();

            winformsMap1.CurrentExtent = rect;
            winformsMap1.Refresh();

            MessageBox.Show(string.Format("Time to display {0} features was {1} seconds",
                db.FeatureSource.GetCount(), s.Elapsed.TotalSeconds));

        }





    }
}
